% a freebody tracking, Tiancheng Li (t.c.li@usal.es)

% Track a body falling through the atmosphere, by EKF/UKF/PF and O2 (observation-only inference) and
% Online Fitting Approach

% This model is taken from [Julier00], which was based on [Athans68].

% For the technical details of the fitting approach as well as the problem model, please refer to the paper:
% T. Li, H. Chen, S. Sun and J.M. Corchado, Joint Smoothing, Tracking, and Forecasting Based
% on Continuous-Time Target Trajectory Fitting, arXiv:1708.02196 [stat.AP]
% or the website: https://sites.google.com/site/tianchengli85/o2

clear 

% addpath C:\Users\lenovo\Documents\MATLAB\ParticleFilter\Resampling\CommonFiles
% addpath C:\Users\lenovo\Documents\MATLAB\ParticleFilter\CommonFiles

global rho Q M a dt dT g

%% two parameters that can be adapted; two different types of system change, one telling estimator, the other not
g = 32.2; % not used in the paper yet, to use by Addgravity
Addgravity = 0; % Use this to control the truth ( =1: add gravity g on the accelaration;; =0:not) without informing any estimators; not used in the paper yet
R = 10^5; % measurement noise variance (ft^2) ;  We can also change this sensor uncertainty, informing all estimators ! 
          % R = 10^4  is used in Julier and Ath's paper; we can also simulate for R = 10^5
% We found: for a large sensor noise R=10^5, EKF/UKF and even PF can be easily cheapted at an atitude of the radar level, namely, x(1)=a!!!
% to avoid this, the O2 and unbiased inference need to be carefully designed
% However, the fitting does not have this problems at all  ===== 
% Fitting is insensitive to the data disorder or target manuvering!!!
%% parameter setting for models, fully the same to Julier00 paper
rho = 5e-5; % lb-sec^2/ft^4
Q = diag([0 0 0]); % process noise covariance
M = 10^5; % horizontal range of position sensor
a = 10^5; % altitude of position sensor
dT = 1; % measurement time step
rksteps = 64;              % RK steps / sampling period
dt = dT/rksteps; % time step for integration (seconds)

T = 50; % simulation length in total (seconds)
ttt = dT : dT : T; % simulation iterating steps

NumStep = numel(ttt);
% Simulate the system. Since Q= 0, the system is deterministic
kk = 0; X_True = zeros(3,NumStep); x = [3e5; -2e4; 1e-3]; % initial true state
    for t = ttt
        kk = kk+1;
        for tau = dt : dt : dT
%            % Fourth order Runge Kutta ingegration
            if  Addgravity
                x = RungeKuttaPlusg(x) + sqrt(dt * Q) * [randn; randn; randn] * dt;
            else
                x = RungeKutta(x) + sqrt(dt * Q) * [randn; randn; randn] * dt;
            end
        end
        X_True(:,kk) = x;
    end
x0 = [3e5; -2e4; 3e-5]; % initial state estimate for filter; same to the paper
P0 = diag([1e6 4e6 10]); % initial estimation error covariance
%% parameter setting for filters
% particle filter
N = 200; % number of particles  

% % UKF 
alpha=1;                                    %default, tunable
ki=0;                                       %default, tunable
beta=2;                                     %default, tunable
lambda=alpha^2*(3+ki)-3;                    %scaling factor
c=3+lambda;                                 %scaling factor
Wm=[lambda/c 0.5/c*ones(1,2*3)];            %weights for means
Wc=Wm;
Wc(1)=Wc(1)+(1-alpha^2+beta);               %weights for covariance
c=sqrt(c);

H_fun = @(x) sqrt(M^2 + (x(1)-a)^2); % observation function

% O2 and unbiased MC debiasing
MCUO2runs =100;

% fitting  
options = optimset('Display','off');
% altirude and velocity fitting are performed separately
A_fun = @(p,tt) sqrt(M^2 + (p(3)*tt.^2+p(2)*tt+p(1)-a).^2); % 
B_fun = @(p,tt) [-p(3)*tt.^2-p(2)*tt-p(1);
                (2*p(3)*tt + p(2))./(p(3)*tt.^2 + p(2)*tt +p(1)).^2]; % 3*p(3)*tt.^2+  p(3)*tt.^3+  -p(3)*tt.^3
% Altirude and velocity fitting are performed jointly:
% AB_fun = @(p,tt) [ sqrt(M^2 + (p(7)*tt.^3+p(3)*tt.^2+p(2)*tt+p(1)-a).^2); % 
%                   -p(6)*tt.^2-p(5)*tt-p(4); % p(6)*dT^3/3+p(5)*dT^2/2+ p(4)*dT;
%                  (2*p(6)*tt +p(5))./(p(6)*tt.^2+p(5)*tt+p(4)).^2]; % /p(8) % -log(-()/rho+a 
%              
% ABC_fun = @(p,tt) [ sqrt(M^2 + (p(7)*tt.^3+p(3)*tt.^2+p(2)*tt+p(1)-a).^2); % 
%                   -p(6)*tt.^2-p(5)*tt-p(4); % p(6)*dT^3/3+p(5)*dT^2/2+ p(4)*dT;
%                  (2*p(6)*tt +p(5))./((p(6)*tt.^2+p(5)*tt+p(4)).^2)/p(8); % 3* p(9)*tt.^2+ -p(3)*tt.^3 
%                  p(8)]; 

%% start monte carlo runs
MSE_PF_M = zeros(3,NumStep); MSE_EKF_M = zeros(3,NumStep);MSE_UKF_M = zeros(3,NumStep);
MSE_fit_M = zeros(3,NumStep); MSE_O2_M = zeros(3,NumStep); MSE_UO2_M = zeros(3,NumStep); 
x_PF = zeros(3,NumStep); x_EKF = zeros(3,NumStep);x_UKF = zeros(3,NumStep);
x_fit = zeros(3,NumStep); x_O2 = zeros(3,NumStep);x_UO2 = zeros(3,NumStep);

Runs = 100;  % total number of MC runs
mc = 0; 
Time_PF_M = zeros(1,Runs); Time_EKF_M = zeros(1,Runs);Time_UKF_M = zeros(1,Runs);
Time_fit_M = zeros(1,Runs); Time_O2_M = zeros(1,Runs); Time_UO2_M = zeros(1,Runs); 

% load freebody ; % uncomment this to reload previous running to continou, if before it was Interrupted 
for mc = mc+1:Runs
    fprintf(['The number of runs:', num2str(mc)]);
%% Simulate observation    
    kk = 0;  % Z = zeros(1,numel(ttt)); Vb = zeros(2,numel(ttt)); 
    Z_True = zeros(1,NumStep);
    for t = ttt
       kk = kk +1;
       % Simulate the noisy measurement.
       z = sqrt(M^2 + (X_True(1,kk)-a)^2) + sqrt(R) * randn;
       Z_True(kk) = z;
    end
    
%% === PF ===  Simulate the continuous-time part of the particle filter (time update).
    % Initialize the PF.
    xpar = zeros(3,N); ErrorP = zeros(1, N);
    for i = 1 : N
        xpar(:,i) = x0 + sqrt(P0) * [randn; randn; randn];
    end

    kk = 0; tic
    for t = ttt
       kk = kk +1; z = Z_True(kk);
       xhatminus = xpar; 
       for i = 1 : N
           for tau = dt : dt : dT
               % Fourth order Runge Kutta ingegration
              xhatminus(:,i) = RungeKutta(xhatminus(:,i))+ sqrt(dt * Q) * [randn; randn; randn] * dt;
              xhatminus(3,i) = max(eps, xhatminus(3,i)); % the ballistic coefficient cannot be negative
           end
           ErrorP(i) = z - sqrt(M^2 + (xhatminus(1,i)-a)^2); % standard particle filter
       end
       % Note that we need to scale all of the q(i) probabilities in a way
       % that does not change their relative magnitudes.
       % Otherwise all of the q(i) elements will be zero because of the
       % large value of the exponential.
       vhatscale = max(abs(ErrorP))/4;
       qsum = 0;
       for i = 1 : N
           q(i) = exp(-(ErrorP(i)/vhatscale)^2); %exp(-(ErrorP(i)^2/2/sqrt(R))); - this standard likelihood does not work!
           qsum = qsum + q(i);
       end
       % Normalize the likelihood of each a priori estimate.
       q = q/qsum;

       % xhat = sum(xhatminus.*[q;q;q],2);
       % Resample.
       [xpar]= resampleRough(q,xhatminus); % Use roughening to prevent sample impoverishment. learn from Gordon

       % The particle filter estimate is the mean of the particles.
       x_PF(:,kk) = mean(xpar')';
       
       MSE_PF_M(:,kk) = MSE_PF_M(:,kk)+ (x_PF(:,kk)-X_True(:,kk)).^2;
    end
    Time_PF_M(mc) = toc;   
%% Observation-only inference
    kk = 0; flag_O2 = 1; tic;
    
    for t = ttt
       kk = kk +1; z = Z_True(kk);
       x_O2(1,kk) = flag_O2*sqrt(z^2-M^2) + a;
%        flag_O2 = sign(X_True(1,kk)-a);  % here sign(x(1)-a) assumes that we know whether the radar face up or down, or use the following adapting way
       if kk==1          
           x_O2([2,3],kk) = x0([2,3]); % we will not really use it  
       else         
           x_O2(2,kk) = (x_O2(1,kk-1) -x_O2(1,kk))/dT; % this, more actually, estimate the velocity at time instant: kk-0.5;                  
                if x_O2(2,kk)<x_O2(2,kk-1) && x_O2(1,kk)<a+x_O2(2,kk-1) && flag_O2
                    % The following two rules if occur jointly, change the sign:
                    % 1. velocity at k is smaller than that at k-1
                    % 2. the sign should only change once in the whole process and only around an atitude of round a
                   flag_O2 = -1;
                   x_O2(1,kk) = flag_O2*sqrt(z^2-M^2)+a;
                   x_O2(2,kk) = x_O2(1,kk-1) -x_O2(1,kk);
                end
           x_O2(3,kk) = (x_O2(2,kk) -x_O2(2,kk-1))/dT/(-1*exp(-rho*x_O2(1,kk)) * x_O2(2,kk)^2);
       end       
       
       MSE_O2_M(:,kk) = MSE_O2_M(:,kk) + (x_O2(:,kk)-X_True(:,kk)).^2;
    end
    Time_O2_M(mc) = toc;      
%% Unbiased Observation-only inference
    kk = 0; flag_UO2 =1; tic
    for t = ttt
       kk = kk +1; z = Z_True(kk);
        x_UO2(1,kk) = mean(flag_UO2*sqrt((z-sqrt(R) * randn(1,MCUO2runs)).^2-M^2))+a;  % MC unbiased projectoion
       if kk==1
           x_UO2([2,3],kk) = x0([2,3]);   
       else
           x_UO2(2,kk) = x_UO2(1,kk-1) -x_UO2(1,kk)/dT;
           if flag_UO2 && x_UO2(2,kk)<x_UO2(2,kk-1) && x_UO2(1,kk)<a+x_UO2(2,kk-1) % The same rules to determine whether to change the sign as used for O2
               flag_UO2 = -1;
               x_UO2(1,kk) = mean(flag_UO2*sqrt((z-sqrt(R) * randn(1,MCUO2runs)).^2-M^2))+a;  % MC unbiased projectoion
               x_UO2(2,kk) = x_UO2(1,kk-1) -x_UO2(1,kk)/dT;
           end
           x_UO2(3,kk) = (x_UO2(2,kk) -x_UO2(2,kk-1))/dT/(-1*exp(-rho*x_UO2(1,kk)) * x_UO2(2,kk)^2 );
       end              
       MSE_UO2_M(:,kk) = MSE_UO2_M(:,kk) + (x_UO2(:,kk)-X_True(:,kk)).^2;    
    end
    Time_UO2_M(mc) = toc; 
%% Fitting
    % Initialize the fitting
    p = [3e5 0.1 0.1 -2e4 0.1 0.1 0.1 3e-5 0.1];  coefficient = 3e-5; % starting parameters are the same to all filters' inilization; but not so important in our case, can be any
    pa = [3e5,0.1,0.1]; pb = [-2e4,0.1,0.1]; 
    % a simple threshold-based protocol for determining the length of sliding time window
    if R >= 10^6
        Window = 10; % when the sensor data is of low quality, we may consider a long time window for fitting
    else
        Window = 5; % ==== whne the sensor data are very good (the trajectory is very smooth), we don't need to fit much! 
    end
    
    kk = 0; tic
    for t = ttt
       kk = kk +1; z = Z_True(kk);
        % individual fitting in altitude, velocity
           flag = max(1,kk-Window);
           pa = lsqcurvefit(A_fun,pa,flag:dT:t,Z_True(flag:kk),[],[],options);
           x_fit(1,kk) = pa(3)*t^2+pa(2)*t+pa(1); %
        % velocity fitting; 
           flag2 = max(1,kk-Window); temp = -exp(-rho*(flag_O2*sqrt(z^2-M^2) + a));  
           Vb(:,kk) = [pa(3)*t*2+pa(2);temp*coefficient];% *x_fit(3,kk)-exp(-rho*x_fit(1,kk))*coefficient  3*pa(3)*t^2+pa(2)*t*2+pa(1)
           pb =lsqcurvefit(B_fun,pb,flag2:dT:t,Vb(:,flag2:kk),[],[],options);
           x_fit(2,kk) = pb(3)*t^2 + pb(2)*t + pb(1); %           
           coefficient = (pb(3)*t*2 + pb(2))/(temp*x_fit(2,kk)^2);coefficient = max(1e-6,coefficient); % -1*exp(-rho*x_fit(1,kk))
           x_fit(3,kk) = coefficient; % mean(BC(1:kk))

%  % joint fitting: do the altitude and velocity fitting jointly  -- not as good as expected
%          Window = 7; flag = max(1,kk-Window);  temp = -exp(-rho*(flag_O2*sqrt(z^2-M^2) + a));     
% %         for jj =1:4 % repeated fitting for better accuracy
%              afit(kk) = 3*p(7)*t^2 + 2*p(3)*t +p(2); %  
%              bfit(kk) = temp*coefficient;% ; % ; % flag_O2*sqrt(z^2-M^2)
%     %          cfit(kk) = p(8); if ABC_fun is used
%                p = lsqcurvefit(AB_fun,p,flag:dT:t,[Z_True(flag:kk); afit(flag:kk); bfit(flag:kk)],[],[],options); % p(8) = max(1e-6,p(8));% ; cfit(flag:kk) if ABC_fun is used
%                x_fit(1,kk) = p(7)*t^3 + p(3)*t^2 + p(2)*t + p(1) ; % 
%                x_fit(2,kk) = p(6)*t^2 + p(5)*t + p(4); % p(9)*t^3 +  
%     %            x_fit(3,kk) = p(8); % coefficient; % if ABC_fun is used
%                coefficient = (2*p(6)*t + p(5))/(temp*x_fit(2,kk)^2); coefficient = max(1e-6,coefficient);
%                x_fit(3,kk) = coefficient;
% % %         end
           
       MSE_fit_M(:,kk) = MSE_fit_M(:,kk) + (x_fit(:,kk)-X_True(:,kk)).^2;
    end
    Time_fit_M(mc) = toc; 
%% EKF   
    % Initialize the EKF 
    x_ekf = x0; P_EKF = P0; 

    kk = 0;tic;
    for t = ttt
       kk = kk +1; z = Z_True(kk);
       % Runge - Kutta time update
        for tau = dt : dt : dT
            [x_ekf,P_EKF] = RungeKuttaKF(x_ekf,P_EKF);
            x_ekf = x_ekf + sqrt(dt * Q) * [randn; randn; randn] * dt;
            x_ekf(3,1) = max(1e-6,x_ekf(3,1));
        end
       H = [(x_ekf(1)-a)/sqrt(M^2+(x_ekf(1)-a)^2) 0 0];
       % Gain updating.
       K = P_EKF * H' * inv(H * P_EKF * H' + R);
       x_ekf = x_ekf + K * (z - sqrt(M^2 + (x_ekf(1)-a)^2)); % H * x_ekf
       P_EKF = (eye(3) - K * H) * P_EKF * (eye(3) - K * H)' + K * R * K';
       x_EKF(:,kk) = x_ekf;

       MSE_EKF_M(:,kk) = MSE_EKF_M(:,kk) + (x_EKF(:,kk)-X_True(:,kk)).^2;
    end
    Time_EKF_M(mc) = toc; 
%% UKF
    % Initialize the UKF 
    x_ukf = x0; P_UKF = P0;
    
    kk = 0; tic;
    for t = ttt
       kk = kk +1; z = Z_True(kk);
       % Runge - Kutta time update
        for tau = dt : dt : dT % sicne the process noise is zero, we don't need the Sigma-step
            [x_ukf,P_UKF] = RungeKuttaKF(x_ukf,P_UKF);
            x_ukf = x_ukf + sqrt(dt * Q) * [randn; randn; randn] * dt;
            x_ukf(3,1) = max(1e-6,x_ukf(3,1));
        end
       % Gain updating.
        A = c*chol((P_UKF+P_UKF')/2)';
        Y = x_ukf(:,ones(1,numel(x_ukf)));
        Xukf = [x_ukf Y+A Y-A]; 

        X2=Xukf-x_ukf(:,ones(1,size(Xukf,2)));             %deviation of X1
        L = size(Xukf,2);
        y = 0;
        LL = zeros(1,L);
        for j=1:L                 
            LL(:,j)=H_fun(Xukf(:,j));       
            y=y+Wm(j)*LL(:,j);       
        end
        Y1=LL-y(:,ones(1,L));
        P2=Y1*diag(Wc)*Y1'+R;  

        P12=X2*diag(Wc)*Y1';                              %transformed cross-covariance
        K=P12/P2;
        x_ukf=x_ukf+K*(z-y);                              %state update
        P_UKF = P_UKF-K*P12';                                %covariance update       
       
       x_UKF(:,kk) = x_ukf;

       MSE_UKF_M(:,kk) = MSE_UKF_M(:,kk) + (x_UKF(:,kk)-X_True(:,kk)).^2;
    end
    Time_UKF_M(mc) = toc; 
    save freebody % to use this to save the results in the middle, in case that some problems happen before all Runs are finished
end

Runs=mc;

RMSE_PF_M = sqrt(MSE_PF_M/Runs);
RMSE_O2_M = sqrt(MSE_O2_M/Runs); 
RMSE_fit_M = sqrt(MSE_fit_M/Runs);
RMSE_EKF_M = sqrt(MSE_EKF_M/Runs);
RMSE_UKF_M = sqrt(MSE_UKF_M/Runs);
RMSE_UO2_M = sqrt(MSE_UO2_M/Runs);

Time_PF_M = mean(Time_PF_M)
Time_O2_M = mean(Time_O2_M)
Time_fit_M = mean(Time_fit_M)
Time_EKF_M = mean(Time_EKF_M)
Time_UKF_M = mean(Time_UKF_M)
Time_UO2_M = mean(Time_UO2_M)

mRMSE_PF_M = mean(RMSE_PF_M,2)
mRMSE_O2_M = mean(RMSE_O2_M,2)
mRMSE_fit_M = mean(RMSE_fit_M,2)
mRMSE_EKF_M = mean(RMSE_EKF_M,2)
mRMSE_UKF_M = mean(RMSE_UKF_M,2)
mRMSE_UO2_M = mean(RMSE_UO2_M,2)

save freebody 

close all;

figure; 
subplot(2,2,1)
semilogy(ttt, X_True(2,:), 'k:',ttt, x_EKF(1,:), 'c-.',ttt, x_UKF(1,:), 'm-.',ttt, x_PF(1,:), 'g--',ttt,  x_O2(1,:), 'b--',ttt,  x_UO2(1,:), 'b.-',ttt, x_fit(1,:), 'r.-'); hold;
set(gca,'FontSize',12); set(gcf,'Color','White');
legend ('Truth','EKF','UKF','PF','O2','Unbiased O2','Fitting');
xlabel('Filtering time /seconds');
ylabel('Altitude');
subplot(2,2,2)
semilogy(ttt, RMSE_EKF_M(1,:), 'c-.',ttt, RMSE_UKF_M(1,:), 'm-.',ttt, RMSE_PF_M(1,:), 'g--',ttt,  RMSE_O2_M(1,:), 'b--',ttt,  RMSE_UO2_M(1,:), 'b.-',ttt, RMSE_fit_M(1,:), 'r.-'); hold;
set(gca,'FontSize',12); set(gcf,'Color','White');
legend ('EKF','UKF','PF','O2','Unbiased O2','Fitting');
xlabel('Filtering time /seconds');
ylabel('Altitude RMSE');
subplot(2,2,3)
semilogy(ttt, RMSE_EKF_M(2,:), 'c-.',ttt, RMSE_UKF_M(2,:), 'm-.',ttt, RMSE_PF_M(2,:), 'g--',ttt,  RMSE_O2_M(2,:), 'b--',ttt,  RMSE_UO2_M(2,:), 'b.-',ttt, RMSE_fit_M(2,:), 'r.-'); hold;
set(gca,'FontSize',12); set(gcf,'Color','White');
legend ('EKF','UKF','PF','O2','Unbiased O2','Fitting');
xlabel('Filtering time /seconds');
ylabel('Velocity RMSE');
subplot(2,2,4)
semilogy(ttt, RMSE_EKF_M(3,:), 'c-.',ttt, RMSE_UKF_M(3,:), 'm-.',ttt, RMSE_PF_M(3,:), 'g--',ttt,  RMSE_O2_M(3,:), 'b--',ttt,  RMSE_UO2_M(3,:), 'b.-',ttt, RMSE_fit_M(3,:), 'r.-'); hold;
set(gca,'FontSize',12); set(gcf,'Color','White');
legend ('EKF','UKF','PF','O2','Unbiased O2','Fitting');
xlabel('Filtering time /seconds');
ylabel('Ballistic Coefficient RMSE');


